from collections import deque
from typing import Deque, List

import numpy as np


def get_entangled_initial_state(initial_state, new_qubit_order):
    states_to_entangle: Deque[np.ndarray] = deque()

    states = list(map(lambda s: np.asarray([1 if s == 0 else 0, 1 if s == 1 else 0], dtype="complex64"), initial_state))

    for q in new_qubit_order:
        states_to_entangle.append(states[q])

    while len(states_to_entangle) > 1:
        entangled_states: Deque[np.ndarray] = deque()
        while len(states_to_entangle) > 1:
            entangled_states.append(np.kron(states_to_entangle.popleft(), states_to_entangle.popleft()))
        if len(states_to_entangle) == 1:
            entangled_states.append(states_to_entangle.popleft())
        states_to_entangle = entangled_states

    return states_to_entangle.pop()


def change_qubit_order(matrix: np.ndarray, old_order, new_order) -> np.ndarray:
    reorder_qubits = [new_order.index(q2) for q2 in old_order]
    initial_row_order = [i for i in range(np.power(2, len(old_order)))]

    reorder_rows = list(map(lambda i: rearrange_bits(i, reorder_qubits), initial_row_order))

    return matrix[reorder_rows][:, reorder_rows]


def rearrange_bits(num: int, new_bit_order: List[int]):
    sig_bits = len(new_bit_order)
    bit_string = int_to_bit_string(num, sig_bits)
    result = int_from_bit_string("".join(bit_string[b] for b in new_bit_order))
    return result


def int_to_bits(num: int, sig_bits: int) -> str:
    return format(num, "#0%sb" % np.power(2, sig_bits))


def int_from_bit_string(bits: str) -> int:
    return int(bits, 2) if bits.startswith("0b") else int("0b" + bits, 2)


def int_to_bit_string(i: int, sig_bits: int) -> str:
    return "{0:b}".format(i).zfill(sig_bits)


def has_common_qbits(gate1, gate2):
    return len(set(gate1.qubits).intersection(set(gate2.qubits))) > 0



import math
import random
from typing import List, Tuple

from shor.gates import CNOT, CSWAP, QFT, H, Rx, X
from shor.layers import Qubits
from shor.quantum import Circuit


def factor(N: int) -> Tuple[int, int]:
    """WIP not currently functioning.

    The classical part seems to be correct, but the quantum period finding function is incorrect
    at the moment.
    """
    while True:
        a = random.randint(2, N - 1)
        common_divisor = gcd(a, N)
        if common_divisor > 1:
            # Lucky guess
            return common_divisor, N / common_divisor

        r = find_period(a, N)

        a_pow_r_o_2 = a ^ r // 2

        if r & 1 == 1 or (a_pow_r_o_2 + 1) % N == 0:
            continue

        factor = max(gcd(a_pow_r_o_2 + 1, N), gcd(a_pow_r_o_2 - 1, N))
        return factor, N // factor


def find_period(a, N):
    """WIP: Quantum subroutine for shor's algorithm.
    Finds the period of a function of the form:
    f(x) = a^x % N

    This uses the quantum fourier transform.
    """

    circuit = Circuit()

    # circuit.add(Qubits(5))
    # circuit.add(QFT(0, 1, 2, 3))
    # circuit.add(X(4))
    # circuit.add(quantum_amod_15(a))
    # circuit.add(QFT(0, 1, 2, 3))

    circuit.add(Qubits(5))
    circuit.add(H(0)).add(H(1)).add(H(2)).add(H(3))
    circuit.add(X(4))
    circuit.add(quantum_amod_15(a))
    circuit.add(QFT(3, 2, 1, 0))  # Inverse Quantum Fourier transform

    job = circuit.run(1024)
    result = job.result

    # TODO: This is still broken, likely due to differences between Qiskit and our library
    # See: https://github.com/shor-team/shor/issues/39
    # from shor.utils.visual import plot_results
    # plot_results(result)

    most_common = result.counts.most_common()
    if len(most_common) > 1:
        return gcd(most_common[0][0], most_common[1][0])
    return 1


def quantum_amod_15(a: int) -> Circuit:
    qc = Circuit()

    if a == 2:
        qc.add(CSWAP(4, 3, 2))
        qc.add(CSWAP(4, 2, 1))
        qc.add(CSWAP(4, 1, 0))
    if a == 4 or a == 11 or a == 14:
        qc.add(CSWAP(4, 2, 0))
        qc.add(CSWAP(4, 3, 1))
        qc.add(CNOT(4, 3))
        qc.add(CNOT(4, 2))
        qc.add(CNOT(4, 1))
        qc.add(CNOT(4, 0))
    if a == 7:
        qc.add(CSWAP(4, 1, 0))
        qc.add(CSWAP(4, 2, 1))
        qc.add(CSWAP(4, 3, 2))
        qc.add(CNOT(4, 3))
        qc.add(CNOT(4, 2))
        qc.add(CNOT(4, 1))
        qc.add(CNOT(4, 0))
    if a == 8:
        qc.add(CSWAP(4, 1, 0))
        qc.add(CSWAP(4, 2, 1))
        qc.add(CSWAP(4, 3, 2))
    if a == 13:
        qc.add(CSWAP(4, 3, 2))
        qc.add(CSWAP(4, 2, 1))
        qc.add(CSWAP(4, 1, 0))
        qc.add(CNOT(4, 3))
        qc.add(CNOT(4, 2))
        qc.add(CNOT(4, 1))
        qc.add(CNOT(4, 0))

    return qc


def qft(qubits: List[int]) -> Circuit:
    qc = Circuit()
    for i in range(len(qubits)):
        for k in range(i):
            qc.add(Rx(qubits[i], qubits[k], angle=math.pi / float(2 ** (i - k))))
        qc.add(H(qubits[i]))
    return qc


def gcd(a: int, b: int) -> int:
    """Finds greatest common divisor between positive integers a and b
    Efficient, using Euclidean algorithm.

    Example:
    >>> greatest_common_divisor(15, 25)
    5
    """
    assert a > 0 and b > 0, "Inputs should be positive integers"

    to_divide, remainder = max(a, b), min(a, b)  # initial divisor
    divisor = 1

    while remainder != 0:
        divisor = remainder
        remainder = to_divide
        remainder = to_divide % divisor
        to_divide = divisor

    return divisor



from abc import ABC, abstractmethod
from enum import Enum
from typing import List

from shor.framework.docstring import _DocExtender
from shor.quantum import QuantumCircuit
from shor.utils.qbits import int_from_bit_string, int_to_bit_string


class Result(ABC):
    @property
    @abstractmethod
    def counts(self):
        pass

    @property
    @abstractmethod
    def sig_bits(self):
        pass

    def get(self, key, default=0):
        if isinstance(key, str):
            idx = int_from_bit_string(key)
        else:
            idx = key

        return self.counts.get(idx, default)

    def __getitem__(self, item):
        return self.get(item)

    def __repr__(self):
        return repr({int_to_bit_string(k, sig_bits=self.sig_bits): v for k, v in self.counts.items()})


class JobStatusCode(Enum):
    COMPLETED = ("completed",)
    ERROR = ("error",)
    RUNNING = ("running",)
    WAITING = "waiting"


class JobStatus(object):
    def __init__(self, code: JobStatusCode, message: str = "", api_error_code: int = None):
        self.code = code
        self.message = message
        self.api_error_code = api_error_code


class Job(ABC):
    @property
    @abstractmethod
    def result(self):
        pass

    @property
    @abstractmethod
    def status(self):
        pass


class Provider(metaclass=_DocExtender):
    """Shor quantum computing providers support:
    - Running shor `QuantumCircuits`
    - Keeping track of running programs (jobs)
    - Listing and selecting devices
    - (Optional) Login and logout

    """

    @abstractmethod
    def devices(self) -> List[str]:
        """List available devices (quantum computers or simulators) for provider
        :return: A list of device names
        """
        pass

    @abstractmethod
    def use_device(self, device: str) -> bool:
        """Choose active device for provider. See `devices()` to list available devices
        :param device: the name of the desired quantum computer or similator
        :return: whether or not operation was successful
        """
        pass

    @abstractmethod
    def run(self, circuit: QuantumCircuit, times: int) -> Job:
        pass

    @property
    @abstractmethod
    def jobs(self) -> List[Job]:
        pass


class WithLoginMixin(metaclass=_DocExtender):
    @abstractmethod
    def account(self) -> str:
        """Shows active account credentials, if logged in."""
        pass

    @abstractmethod
    def login(self, token: str, remember: bool = False, **kwargs) -> bool:
        """Login to provider with API token.

        :param token: API token
        :param remember: Flag to save credentials to file system for future use
        :param kwargs: Additional params to pass to the downstream provider login function
        :return: boolean if login was successful
        """
        pass

    @abstractmethod
    def logout(self, forget: bool = False) -> None:
        """Logout of provider.

        :param forget: Param to forget saved credentials
        """
        pass



import math
from typing import Iterable, Union

import numpy as np

from shor.errors import CircuitError
from shor.layers import _Layer
from shor.utils.collections import flatten

QbitOrIterable = Union[int, Iterable]


class _Gate(_Layer):
    """Abstract base quantum gate class

    # Properties
    input_length = valid length of input qubits
    qubits = indices of qubits, to be used as input to gate.
    """

    @property
    def symbol(self):
        return self.__class__.__name__.lower()

    def __init__(self, *qbits: QbitOrIterable, **kwargs):
        super().__init__(**kwargs)
        self.qbits = flatten(qbits) if qbits else [0]
        self.dimension = kwargs.get("dimension", 1)

        assert all(map(lambda q: type(q) == int, self.qbits)), str(self.qbits)
        try:
            assert len(self.qbits) % self.dimension == 0
        except AssertionError:
            raise CircuitError(
                f"The input qbits length {len(self.qbits)} is not divisible by the '{self.symbol}' "
                f"gate's dimension {self.dimension}"
            )

    @property
    def qubits(self):
        return self.qbits

    def to_gates(self):
        if len(self.qbits) > self.dimension:
            return [
                self.__class__(self.qbits[i : i + self.dimension]) for i in range(0, len(self.qbits), self.dimension)
            ]
        return [self]

    @property
    def num_states(self):
        return np.power(2, self.dimension)

    def to_matrix(self) -> np.ndarray:
        return np.eye(self.num_states())

    @property
    def matrix(self):
        return self.to_matrix()

    def invert(self):
        return self

    @property
    def I(self):
        return self.invert()

    def __invert__(self):
        return self.invert()


class CNOT(_Gate):
    symbol = "CX"

    def __init__(self, *qubits, **kwargs):
        kwargs["dimension"] = 2
        if not qubits:
            qubits = [0, 1]

        super().__init__(*qubits, **kwargs)

    @staticmethod
    def to_matrix() -> np.ndarray:
        return np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]])


class CY(_Gate):
    symbol = "CY"

    def __init__(self, *qubits, **kwargs):
        kwargs["dimension"] = 2
        if not qubits:
            qubits = [0, 1]

        super().__init__(*qubits, **kwargs)

    @staticmethod
    def to_matrix() -> np.ndarray:
        return np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, -1 * 1j], [0, 0, 1j, 0]])


class CSWAP(_Gate):
    symbol = "CSWAP"

    def __init__(self, *qubits, **kwargs):
        kwargs["dimension"] = 3
        if not qubits:
            qubits = [0, 1, 2]

        super().__init__(*qubits, **kwargs)

    @staticmethod
    def to_matrix() -> np.ndarray:
        cswap_matrix = np.eye(8)
        cswap_matrix[:, [5, 6]] = cswap_matrix[:, [6, 5]]
        return cswap_matrix


class Hadamard(_Gate):
    symbol = "H"

    def __init__(self, *qubits, **kwargs):
        kwargs["dimension"] = 1
        if not qubits:
            qubits = [0]

        super().__init__(*qubits, **kwargs)

    def to_matrix(self) -> np.ndarray:
        return np.multiply(np.divide(1, np.sqrt(self.num_states)), np.array([[1, 1], [1, -1]]))


class PauliX(_Gate):
    symbol = "X"

    def __init__(self, *qubits, **kwargs):
        kwargs["dimension"] = 1
        if not qubits:
            qubits = [0]

        super().__init__(*qubits, **kwargs)

    @staticmethod
    def to_matrix() -> np.ndarray:
        return np.array([[0, 1], [1, 0]])


class PauliY(_Gate):
    symbol = "Y"

    def __init__(self, *qubits, **kwargs):
        kwargs["dimension"] = 1
        if not qubits:
            qubits = [0]

        super().__init__(*qubits, **kwargs)

    @staticmethod
    def to_matrix() -> np.ndarray:
        return np.array([[0, -1j], [1j, 0]])


class PauliZ(_Gate):
    symbol = "Z"

    def __init__(self, *qubits, **kwargs):
        kwargs["dimension"] = 1
        if not qubits:
            qubits = [0]

        super().__init__(*qubits, **kwargs)

    @staticmethod
    def to_matrix() -> np.ndarray:
        return np.array([[1, 0], [0, -1]])


class QFT(_Gate):
    def __init__(self, *qubits, **kwargs):
        if not qubits:
            qubits = [0, 1]

        super().__init__(*qubits, dimension=len(qubits), **kwargs)

    # def to_gates(self):
    #     # TODO: translate this gate to base gates / CNOTs
    #     pass

    def get_nth_unity_root(self, k):
        return np.exp((2j * np.pi * k) / self.num_states)

    def to_matrix(self) -> np.ndarray:
        m = np.array(np.ones((self.num_states, self.num_states)), dtype="complex")

        for i in range(1, self.num_states):
            for j in range(i, self.num_states):
                w = self.get_nth_unity_root(i * j)
                m[i, j] = w
                m[j, i] = w

        return np.around(np.multiply(1 / np.sqrt(self.num_states), m), decimals=15)


class SWAP(_Gate):
    symbol = "SWAP"

    def __init__(self, *qubits, **kwargs):
        kwargs["dimension"] = 2
        if not qubits:
            qubits = [0, 1]

        super().__init__(*qubits, **kwargs)

    @staticmethod
    def to_matrix() -> np.ndarray:
        return np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])


class Cx(_Gate):
    symbol = "CX"

    def __init__(self, *qubits, **kwargs):
        kwargs["dimension"] = 2
        if not qubits:
            qubits = [0, 1]

        super().__init__(*qubits, **kwargs)

    @staticmethod
    def to_matrix() -> np.ndarray:
        return np.array([[1, 0, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0], [0, 1, 0, 0]])


class CCNOT(_Gate):
    symbol = "CCX"

    def __init__(self, *qubits, **kwargs):
        kwargs["dimension"] = 3
        if not qubits:
            qubits = [0, 1, 2]

        super().__init__(*qubits, **kwargs)

    @staticmethod
    def to_matrix() -> np.ndarray:
        return np.array(
            [
                [1, 0, 0, 0, 0, 0, 0, 0],
                [0, 1, 0, 0, 0, 0, 0, 0],
                [0, 0, 1, 0, 0, 0, 0, 0],
                [0, 0, 0, 1, 0, 0, 0, 0],
                [0, 0, 0, 0, 1, 0, 0, 0],
                [0, 0, 0, 0, 0, 1, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 1],
                [0, 0, 0, 0, 0, 0, 1, 0],
            ]
        )


class CRZ(_Gate):
    symbol = "CRZ"

    def __init__(self, *qubits, angle=0, **kwargs):
        kwargs["dimension"] = 2
        self.angle = angle
        if not qubits:
            qubits = [0, 1]

        super().__init__(*qubits, **kwargs)

    def to_matrix(self) -> np.ndarray:
        return np.array(
            [
                [1, 0, 0, 0],
                [0, 1, 0, 0],
                [0, 0, np.exp(-1j * self.angle / 2), 0],
                [0, 0, 0, np.exp(1j * self.angle / 2)],
            ]
        )


class CH(_Gate):
    symbol = "CH"

    def __init__(self, *qubits, **kwargs):
        kwargs["dimension"] = 2
        if not qubits:
            qubits = [0, 1]

        super().__init__(*qubits, **kwargs)

    @staticmethod
    def to_matrix() -> np.ndarray:
        return np.array(
            [
                [1, 0, 0, 0],
                [0, 1, 0, 0],
                [0, 0, 1 / np.sqrt(2), 1 / np.sqrt(2)],
                [0, 0, 1 / np.sqrt(2), -1 / np.sqrt(2)],
            ]
        )


class S(_Gate):
    symbol = "S"

    def __init__(self, *qubits, **kwargs):
        kwargs["dimension"] = 1
        if not qubits:
            qubits = [0]

        super().__init__(*qubits, **kwargs)

    @staticmethod
    def to_matrix() -> np.ndarray:
        return np.array([[1, 0], [0, 1j]])


class Sdg(_Gate):
    symbol = "Sdg"

    def __init__(self, *qubits, **kwargs):
        kwargs["dimension"] = 1
        if not qubits:
            qubits = [0]

        super().__init__(*qubits, **kwargs)

    @staticmethod
    def to_matrix() -> np.ndarray:
        return np.array([[1, 0], [0, -1j]])


class T(_Gate):
    symbol = "T"

    def __init__(self, *qubits, **kwargs):
        kwargs["dimension"] = 1
        if not qubits:
            qubits = [0]

        super().__init__(*qubits, **kwargs)

    @staticmethod
    def to_matrix() -> np.ndarray:
        return np.array([[1, 0], [0, np.exp(1j * np.pi / 4)]])


class Tdg(_Gate):
    symbol = "Tdg"

    def __init__(self, *qubits, **kwargs):
        kwargs["dimension"] = 1
        if not qubits:
            qubits = [0]

        super().__init__(*qubits, **kwargs)

    @staticmethod
    def to_matrix() -> np.ndarray:
        return np.array([[1, 0], [0, np.exp(-1j * np.pi / 4)]])


class ID(_Gate):
    symbol = "I"

    def __init__(self, *qubits, **kwargs):
        kwargs["dimension"] = 1
        if not qubits:
            qubits = [0]

        super().__init__(*qubits, **kwargs)

    @staticmethod
    def to_matrix() -> np.ndarray:
        return np.array([[1, 0], [0, 1]])


class U1(_Gate):
    symbol = "U1"

    def __init__(self, *qubits, angle=0, **kwargs):
        kwargs["dimension"] = 1
        self.angle = angle
        if not qubits:
            qubits = [0]

        super().__init__(*qubits, **kwargs)

    def to_matrix(self) -> np.ndarray:
        return np.array([[1, 0], [0, np.exp(1j * self.angle)]])


class Cz(_Gate):
    symbol = "CZ"

    def __init__(self, *qubits, **kwargs):
        kwargs["dimension"] = 2
        if not qubits:
            qubits = [0, 1]

        super().__init__(*qubits, **kwargs)

    @staticmethod
    def to_matrix() -> np.ndarray:
        return np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, -1]])


class Rx(_Gate):
    symbol = "RX"

    def __init__(self, *qubits, angle=math.pi / 2, **kwargs):
        kwargs["dimension"] = 1
        self.angle = angle
        if not qubits:
            qubits = [0]

        super().__init__(*qubits, **kwargs)

    def to_matrix(self) -> np.ndarray:
        return np.array(
            [
                [math.cos(self.angle / 2), -math.sin(self.angle / 2) * 1j],
                [-math.sin(self.angle / 2) * 1j, math.cos(self.angle / 2)],
            ]
        )


class Ry(_Gate):
    symbol = "RY"

    def __init__(self, *qubits, angle=math.pi / 2, **kwargs):
        kwargs["dimension"] = 1
        self.angle = angle
        if not qubits:
            qubits = [0]

        super().__init__(*qubits, **kwargs)

    def to_matrix(self) -> np.ndarray:
        return np.array(
            [
                [math.cos(self.angle / 2), -math.sin(self.angle / 2)],
                [math.sin(self.angle / 2), math.cos(self.angle / 2)],
            ]
        )


class Rz(_Gate):
    symbol = "RZ"

    def __init__(self, *qubits, angle, **kwargs):
        self.angle = angle
        kwargs["dimension"] = 1
        if not qubits:
            qubits = [0]
        super().__init__(*qubits, **kwargs)

    def to_matrix(self) -> np.ndarray:
        return np.array(
            [[np.exp(-(1 / 2) * 1j * self.angle), 0], [0, np.exp((1 / 2) * 1j * self.angle)]], dtype="complex"
        )


class U3(_Gate):
    symbol = "U3"

    def __init__(self, *qubits, theta=0, phi=0, lam=0, **kwargs):
        kwargs["dimension"] = 1
        self.theta = theta
        self.phi = phi
        self.lam = lam

        if not qubits:
            qubits = [0]

        super().__init__(*qubits, **kwargs)

    def to_matrix(self) -> np.ndarray:
        return np.array(
            [
                [math.cos(self.theta / 2), -np.exp(1j * self.lam) * math.sin(self.theta / 2)],
                [
                    np.exp(1j * self.phi) * math.sin(self.theta / 2),
                    np.exp(1j * (self.phi + self.lam)) * math.cos(self.theta / 2),
                ],
            ]
        )


class U2(U3):
    symbol = "U2"

    def __init__(self, *qubits, phi=0, lam=0, **kwargs):
        super().__init__(*qubits, theta=np.pi / 2, phi=phi, lam=lam, **kwargs)
        self.symbol = "u2"


class Init_x(_Gate):
    def __init__(self, *qubits, **kwargs):
        self.H = Hadamard(0)
        kwargs["dimension"] = 1
        super().__init__(*qubits, **kwargs)

    def to_matrix(self) -> np.ndarray:
        return self.H.to_matrix()


class Init_y(_Gate):
    def __init__(self, *qubits, **kwargs):
        self.H = Hadamard(0)
        self.S = S()
        kwargs["dimension"] = 1
        super().__init__(*qubits, **kwargs)

    def to_matrix(self) -> np.ndarray:
        return self.S.to_matrix().dot(self.H.to_matrix())


class Cr(_Gate):
    symbol = "CU1"

    def __init__(self, *qubits, angle, **kwargs):
        self.angle = angle
        kwargs["dimension"] = 2
        if not qubits:
            qubits = [0]
        super().__init__(*qubits, **kwargs)

    def to_matrix(self) -> np.ndarray:
        return np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, np.exp(1j * self.angle)]], dtype="complex")


class CRk(_Gate):
    def __init__(self, *qubits, k, **kwargs):
        self.k = k
        kwargs["dimension"] = 2
        if not qubits:
            qubits = [0]
        super().__init__(*qubits, **kwargs)

    def to_matrix(self) -> np.ndarray:
        return np.array(
            [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, np.exp(2 * 1j * np.pi / 2 ** self.k)]], dtype="complex"
        )


# Aliases
H = h = Hadamard
X = x = PauliX
Y = y = PauliY
Z = z = PauliZ
swap = SWAP
Fredkin = cswap = CSWAP
CX = cx = CNOT



from qiskit.providers.ibmq import IBMQ as QiskitIBMQ
from qiskit.providers.ibmq import IBMQAccountCredentialsNotFound

from shor.framework.docstring import _Hints, doc
from shor.providers import Aer
from shor.providers.base import WithLoginMixin
from shor.providers.qiskit.base import QiskitProvider

DEFAULT_DEVICE = "qasm_simulator"


class _IBMQHints(_Hints):
    LOGIN = """
To login and remember your IBMQ account use `IBMQ.login(TOKEN, remember=True)`
Your TOKEN can be found at: https://quantum-computing.ibm.com/account
"""


class IBMQ(QiskitProvider, WithLoginMixin):
    __doc__ = f"""{{pdoc}}
    The IBMQ provider enables access to IBM quantum computers.

    An IBMQ Experience account is required.
    If no account is provided, `{DEFAULT_DEVICE}` will be chosen as the device

    {_IBMQHints.LOGIN.indent(4)}
    """

    def __init__(self, device=DEFAULT_DEVICE, **config):
        """
        Initialize an IBMQ provider. Uses Qiskit to run programs on IBMQ quantum computers.

        If you need to access the Qiskit provider directly for any reason, use `.provider`

        :param device: Initialize with a specific device, defaults to qasm_simulator. See `devices` to list options
        :param config: Additional Qiskit provider configuration. Passed along as key word arguments to Qiskit's IBMQ provider
        """

        IBMQ._try_login_saved_account()
        provider = QiskitIBMQ.providers()[0] if QiskitIBMQ.providers() else Aer

        config["provider"] = provider

        if "backend" in config:
            if not isinstance(config["backend"], str):
                config["backend"] = provider.get_backend(DEFAULT_DEVICE) if provider and provider.backends() else None

        super().__init__(**config)

    @classmethod
    def account(cls) -> str:
        active_account = QiskitIBMQ.active_account()
        if not active_account:
            print("No active account")
        return active_account if active_account else ""

    @classmethod
    @doc(
        f"""{{pdoc}}

    {_IBMQHints.LOGIN}
"""
    )
    def login(cls, token: str = "", remember: bool = False, **kwargs) -> bool:

        if not token:
            return cls._try_login_saved_account()
        else:
            QiskitIBMQ.enable_account(token, **kwargs)
            if remember:
                QiskitIBMQ.save_account(token, **kwargs)

    @classmethod
    def logout(cls, forget: bool = False) -> None:
        if cls.account():
            if forget:
                QiskitIBMQ.delete_account()
            QiskitIBMQ.disable_account()

    @classmethod
    def _try_login_saved_account(cls) -> bool:
        if not QiskitIBMQ.active_account():
            try:
                QiskitIBMQ.load_account()
            except IBMQAccountCredentialsNotFound:
                print(f"No saved account found for IBMQ. Only simulator devices will be available. {_IBMQHints.LOGIN}")
                return False  # Failed login
        return True  # Successful login



from typing import List, Union

from qiskit import Aer, execute
from qiskit.providers import BaseBackend, BaseProvider

from shor.errors import ProviderError
from shor.providers.base import Job, Provider, Result
from shor.quantum import QC
from shor.transpilers.qiskit import to_qiskit_circuit
from shor.utils.qbits import int_from_bit_string

DEFAULT_DEVICE = Aer.get_backend("qasm_simulator")
DEFAULT_PROVIDER = Aer


class QiskitResult(Result):
    def __init__(self, qiskit_result):
        self.qiskit_result = qiskit_result

    @property
    def counts(self):
        return {int_from_bit_string(k.split(" ")[0]): v for k, v in self.qiskit_result.get_counts().items()}

    @property
    def sig_bits(self):
        measurement_bases = list(self.qiskit_result.get_counts().keys())
        return len(measurement_bases[0]) if measurement_bases else 0


class QiskitJob(Job):
    def __init__(self, qiskit_job):
        self.qiskit_job = qiskit_job

    @property
    def status(self):
        return self.qiskit_job.status()

    @property
    def result(self) -> QiskitResult:
        return QiskitResult(self.qiskit_job.result())


class QiskitProvider(Provider):
    def __init__(self, device=None, provider=DEFAULT_PROVIDER, **config):

        if not issubclass(provider.__class__, BaseProvider):
            raise ProviderError(
                f"Qiskit Provider improperly initialized - must be a subclass of "
                f"<Qiskit.providers.BaseProvider>. The provided provider is not: {provider}"
            )

        self.provider = provider

        if device:
            self.use_device(device)
        elif "backend" in config:
            # Qiskit uses the language "backend" instead of device, attempt to load this as well
            self.use_device(config["backend"])
        else:
            self.device = config.get("backend", DEFAULT_DEVICE)

    def devices(self, **kwargs) -> List[str]:
        return [d.name for d in self.provider.backends(**kwargs)]

    def use_device(self, device: Union[str, BaseBackend], **kwargs) -> bool:
        if type(device) is str:
            self.device = self.provider.get_backend(name=device, **kwargs)
        elif issubclass(device.__class__, BaseBackend):
            self.device = device

        return self.device is not None

    def run(self, circuit: QC, times: int) -> QiskitJob:
        job = execute(to_qiskit_circuit(circuit), self.device, shots=times)

        return QiskitJob(job)

    @property
    def jobs(self) -> List[Job]:
        return list(map(lambda j: QiskitJob(j), self.device.get_jobs()))



from qiskit.providers.aer import Aer as QiskitAer

from shor.providers.qiskit.base import QiskitProvider

DEFAULT_BACKEND = QiskitAer.get_backend("qasm_simulator")


class Aer(QiskitProvider):
    def __init__(self, **config):
        config["backend"] = config.get("backend", DEFAULT_BACKEND)
        config["provider"] = QiskitAer
        super().__init__(**config)



from abc import ABCMeta
from enum import Enum
from inspect import getdoc
from textwrap import indent


class _Hints(Enum):
    def indent(self, n):
        return indent(self.value, prefix=" " * n)

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value


class _DocExtender(ABCMeta):
    def __new__(cls, name, bases, spec):
        for key, value in spec.items():
            if type(value) is classmethod:
                value = value.__func__

            doc = getattr(value, "__doc__", "{pdoc}")

            pdoc = ""
            for b in bases:
                try:
                    pdoc = getdoc(getattr(b, key))
                except (IndexError, AttributeError):
                    pdoc = ""
                if pdoc:
                    break
            try:
                value.__doc__ = doc.format(pdoc=pdoc)
            except AttributeError:
                pass

        if bases:
            pdoc = getattr(bases[0], "__doc__")
            if "__doc__" in spec and spec["__doc__"]:
                spec["__doc__"] = spec["__doc__"].format(pdoc=pdoc)
            else:
                spec["__doc__"] = pdoc

        return ABCMeta.__new__(cls, name, bases, spec)


def doc(value):
    def _doc(func):
        func.__doc__ = value
        return func

    return _doc



from .base import Job, Provider, Result
from .qiskit.Aer import Aer
from .qiskit.base import QiskitJob, QiskitProvider, QiskitResult
from .qiskit.ibmq import IBMQ

__all__ = [Job, Provider, Result, QiskitJob, QiskitResult, QiskitProvider, Aer, IBMQ]



import matplotlib.pyplot as plt

from shor.providers.base import Result
from shor.quantum import Circuit

# import pydot


def plot_circuit(circuit: Circuit, to_file="model.png"):
    """Converts a Quantum Circuit into a diagram, and saves to a file

    Args:
        circuit: A quantum :func:`Circuit() <shor.quantum.Circuit>` object
        to_file: File name of the plot image.

    Returns:
        A Jupyter notebook Image object if Jupyter is installed.
        This enables in-line display of the circuit diagram in notebooks.
    """
    pass


def circuit_to_dot(circuit: Circuit):
    """Converts a Quantum Circuit into Dot format
    See: pydot

    Args:
        circuit: A shor.quantum.Circuit object
    """
    pass


def plot_results(result: Result):
    plt.bar(list(result.counts.keys()), result.counts.values())
    plt.xticks([i for i in range(0, 2 ** result.sig_bits, 5)])
    plt.show()



from typing import List, Union

import numpy as np

from shor.errors import CircuitError
from shor.gates import _Gate
from shor.layers import Qbits, _Layer
from shor.operations import Measure, _Operation


class QuantumCircuit(object):
    def __init__(self):
        self.layers: List[_Layer] = []

    def add(self, layer_or_circuit: Union[_Layer, "QuantumCircuit"]):
        if isinstance(layer_or_circuit, _Layer):
            self.layers.append(layer_or_circuit)
        elif isinstance(layer_or_circuit, QuantumCircuit):
            self.layers.extend(layer_or_circuit.layers)
        else:
            raise TypeError("QuantumCircuit class cannot add the type: {}".format(type(layer_or_circuit)))

        return self

    def draw(self):
        # Use the qiskit drawing function, may want to replace in future.
        from shor.transpilers.qiskit import to_qiskit_circuit

        return to_qiskit_circuit(self).draw()

    def initial_state(self) -> np.ndarray:
        initial_qubits = []
        for qbit_layer in filter(lambda layer: type(layer) == Qbits, self.layers):
            initial_qubits.extend([qbit_layer.state] * qbit_layer.num)

        return initial_qubits

    def to_gates(self, include_operations=False) -> List[_Gate]:
        gates = []
        for layer in self.layers:
            operation_or_gates = [layer] if include_operations and isinstance(layer, _Operation) else layer.to_gates()

            gates.extend(operation_or_gates)
        return gates

    def measure_bits(self):
        measure_bits = []
        for m in filter(lambda l: type(l) == Measure, self.layers):
            measure_bits.extend(m.qbits)

        if not measure_bits:
            raise CircuitError("No measurement found. Valid quantum circuits must contain a 'Measurement' operation")

        return measure_bits

    def __add__(self, other):
        return self.add(other)

    def run(self, num_shots: int, provider=None, **kwargs):
        if provider is None:
            from shor.providers import Aer

            provider = Aer()

        return provider.run(self, num_shots)


# Aliases
Circuit = QC = QuantumCircuit



from collections import Iterable

from shor.errors import CircuitError
from shor.layers import _Layer


class _Operation(_Layer):
    """Abstract base quantum computing operation class

    Operations can be, but are not in general, pure quantum transformations (Unitary transformations)

    The generalized operaation is any legal quantum computer operation, which can include:
    - Unitary operations
    - Measurement
    - Logic conditioned on measurement
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def to_gates(self):
        return []


class Measure(_Operation):
    symbol = "measure"

    def __init__(self, *qbits, output_bits=None, axis="z", **kwargs):
        if not qbits:
            qbits = (0,)

        if len(qbits) == 1 and isinstance(qbits[0], Iterable):
            qbits = tuple(qbits[0])

        if output_bits:
            bits = output_bits
        else:
            bits = qbits[:]

        self.qbits = qbits
        self.bits = bits

        if len(bits) > len(qbits):
            raise CircuitError("Attempting to measure more bits than there are qbits.")

        super().__init__(**kwargs, axis=axis)


# Aliases
M = m = Measure



from collections import Iterable


class _Layer(object):
    """Abstract base quantum layer class"""

    def __init__(self, **kwargs):
        self.name = kwargs.get("name", "Layer")
        pass

    def to_gates(self):
        pass


class Qbits(_Layer, Iterable):
    def __init__(self, num, state=0, **kwargs):
        self.num = num
        self.state = state
        self._qbits = list(range(num))

        super().__init__(name="Qbits ({})".format(str(num)), **kwargs)

    def to_gates(self):
        return []

    def __iter__(self):
        return self._qbits.__iter__()

    def __getitem__(self, key):
        return self._qbits[key]


# Aliases
Qubits = Qbits



from qiskit import QuantumCircuit

from shor.layers import _Layer
from shor.operations import _Operation
from shor.quantum import QC


def to_qiskit_circuit(shor_circuit: QC) -> QuantumCircuit:
    qiskit_circuit = QuantumCircuit(
        len(shor_circuit.initial_state()),
        len(shor_circuit.measure_bits()),
    )

    for gate_or_op in shor_circuit.to_gates(include_operations=True):
        transpile_gate(qiskit_circuit, gate_or_op)

    return qiskit_circuit


def to_qiskit_symbol(shor_layer: _Layer) -> str:
    return shor_layer.symbol.lower()


def transpile_gate(qiskit_circuit: QuantumCircuit, shor_layer: _Layer):
    symbol = to_qiskit_symbol(shor_layer)

    if isinstance(shor_layer, _Operation):
        if symbol == "measure":
            if len(shor_layer.qbits) == len(shor_layer.bits) == qiskit_circuit.size():
                qiskit_circuit.measure_active()
            else:
                for qbit, bit in zip(shor_layer.qbits, shor_layer.bits):
                    qiskit_circuit.measure(qbit, bit)

        return

    args = []
    kwargs = {}

    if symbol in ["crx", "cry", "crz", "cu1", "rx", "ry", "rz", "u1"]:
        args.append(shor_layer.angle)
    if symbol in ["u", "u3"]:
        args.append(shor_layer.theta)
    if symbol in ["u", "u2", "u3"]:
        args.append(shor_layer.phi)
        args.append(shor_layer.lam)

    args.extend(shor_layer.qbits)
    qiskit_circuit.__getattribute__(symbol)(*args, **kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['shor',
 'shor.algorithms',
 'shor.framework',
 'shor.providers',
 'shor.providers.qiskit',
 'shor.transpilers',
 'shor.utils']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.3.3,<4.0.0', 'numpy>=1.19.4,<2.0.0', 'qiskit>=0.23.1,<0.24.0']

setup_kwargs = {
    'name': 'shor',
    'version': '0.0.6',
    'description': 'Quantum Computing for Humans',
    'long_description': '# shor\nThe Python quantum computing library.\n\nShor is a simple, modular, and extensible quantum computing API written in python.\n\n#### Our mission is to make quantum computing easy and accesible to everyone.\n\nUse shor if you need a library that:\n- Enables fast prototyping of quantum circuits\n- Simulates small quantum programs\n- Makes quantum computing easy\n\nWe currently support running programs on Qiskit providers including IBMQ quantum computeres.\n\nIn the near future we plan to support other providers. Write your program once, and run it anywhere.\n\nRead the documentation, view tutorials and examples, and join the community at: https://shor.dev/\n\nWe are in early development and looking for contributors!\n\nWant to contribute? See [CONTRIBUTING.md](CONTRIBUTING.md)\n\nShor is compatible with Python 3.6 and is distributed under the MIT license.',
    'author': 'Collin Overbay - shor.dev',
    'author_email': 'shordotdev@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)
